local UH = UnderHood
local UHB = UnderHood_Bars
local L = LibStub( "AceLocale-3.0" ):GetLocale( "UnderHood" )

local metaStyles = {
	VerticalLarge = {
		title = L["Vertical (Large)"],
	    config = {
	        width = 30,
	        height = 256,
		},
		styleSettings = {
			showBackground = true,
			bgColor = { 0.3, 0.3, 0.3, 0.75 },
		},
	    elements = {
			background = {
			    texture = "Interface\\Addons\\UnderHood\\Media\\Bar-Large",
			    layer = "BACKGROUND",
			    blend = "BLEND",
			    points = true,
			    color = { 0.3, 0.3, 0.3, 0.75 },
			},
			border = {
			    texture = "Interface\\Addons\\UnderHood\\Media\\Border-Large",
			    points = true,
			    blend = "BLEND",
				layer = "BORDER",
			},
			statusbar = {
			    texture = "Interface\\Addons\\UnderHood\\Media\\Bar-Large",
			    points = {
					{ "BOTTOM", "BOTTOM", 0, 3 },
					{ "LEFT", "LEFT", 0, 0 },
					{ "RIGHT", "RIGHT", 0, 0 },
				},
			    blend = "BLEND",
				gaps = { 3, 3, 0, 0 },
				layer = "ARTWORK",
				useBarColor = true,
			},
			castingbar = {
			    texture = "Interface\\Addons\\UnderHood\\Media\\CastingBar-Large",
			    points = {
					{ "BOTTOM", "BOTTOM", 0, 1 },
					{ "LEFT", "LEFT", 0, 0 },
					{ "RIGHT", "RIGHT", 0, 0 },
				},
			    blend = "BLEND",
				gaps = { 1, 1, 0, 0 },
				layer = "ARTWORK",
				useCastingColor = true,
			},
		},
	},
}

local Factory
local Style
do
	Style = {}
	Style.__index = Style
	
	function Style:IsCastingSupported()
	    return self.bar.frames.castingbar
	end
	
	function Style:AdjustBar( element, p )
	    local data = self.meta.elements[element]
		local gapTop, gapBottom, gapLeft, gapRight = 0, 0, 0, 0
		
		if data.gaps then
		    gapTop, gapBottom, gapLeft, gapRight = unpack( data.gaps )
		end
		
		local gapTopP = gapTop / self.bar.config.height
		local gapBottomP = gapBottom / self.bar.config.height
		local h = (self.bar.config.height - gapTop - gapBottom) * p
		local top = 1 - (p - gapTopP)
		local bottom = 1 - gapBottomP
		
		top = top - (gapTopP + gapBottomP) * (1 - p)
		
		self.bar.frames[element]:SetHeight( h )
		self.bar.frames[element]:SetTexCoord( 0, 1, top, bottom )
		
		if math.floor( p * 100 ) == 0 then
			self.bar.frames[element]:Hide()
		else
			self.bar.frames[element]:Show()
		end
	end
	
	function Style:CreateMetaElement( element )
	    local data = self.meta.elements[element]
	    local texture = UH:CreateFrame( "Texture", self.bar.frame, data.layer )
	    
	    texture:SetTexture( data.texture )
	    
	    if data.color then
	        local r, g, b, a = unpack( data.color )
	        
	        texture:SetVertexColor( r, g, b, a )
	    end
	    
	    if data.blend then
	        texture:SetBlendMode( data.blend )
		end
		
	    if data.points then
	        if type( data.points ) == "table" then
	            for _, point in ipairs( data.points ) do
	                local pt, rpt, x, y = unpack( point )
	                
	                texture:SetPoint( pt, self.bar.frame, rpt, x, y )
	            end
	        else
	            texture:SetAllPoints( self.bar.frame )
	        end
	    end
	    
	    if not( data.hidden ) then
	        texture:Show()
		end
		
	    self.bar.frames[element] = texture
	    
	    return texture
	end
	
	function Style:Create( bar, meta )
		local self = {}
		setmetatable( self, Style )
		
		self.bar = bar
		self.meta = meta
		
		if not bar.frames then bar.frames = {} end

		for element, _ in pairs( meta.elements ) do
		    self:CreateMetaElement( element )
		end
		
		if bar.frames.background then
			if bar.config.styleSettings.showBackground then
				bar.frames.background:Show()
			else
				bar.frames.background:Hide()
			end
		end
		
		self:AdjustBar( "statusbar", 1 )
		
		if self.bar.frames.castingbar then
			self:AdjustBar( "castingbar", 0 )
		end
		
		return self
	end
	
	function Style:Destroy( bar )
	end
	
	function Style:SetColor( r, g, b, a )
		for en, ed in pairs( self.meta.elements ) do
		    if ed.useBarColor and self.bar.frames[en] then
		        self.bar.frames[en]:SetVertexColor( r, g, b, a )
		    end
		end
	end

	function Style:SetCastingColor( r, g, b, a )
		for en, ed in pairs( self.meta.elements ) do
		    if ed.useCastingColor and self.bar.frames[en] then
		        self.bar.frames[en]:SetVertexColor( r, g, b, a )
		    end
		end
	end
	
	function Style:SetValue( value )
		self.value = value

		self:Update()		
	end
	
	function Style:SetCastingValue( value )
	    self.castingValue = value
	    
	    self:Update()
	end
	
	function Style:Update()
		local vp, cp = 0, 0
		
		if self.bar:InConfigMode() then
			vp = 1
		else
			vp = self.value or 0
			cp = self.castingValue or 0
		end
		
		if vp > 1 then vp = 1 end
		if cp > 1 then cp = 1 end

		self:AdjustBar( "statusbar", vp )
		
		if self.bar.frames.castingbar then
			self:AdjustBar( "castingbar", cp )
		end
	end
	
	function Style:SetupOptions( options )
		options.showBackground = {
			name = L["Show background"],
			type = "toggle",
			get = function() return self.bar.config.styleSettings.showBackground end,
			set = function( info, value )
				self.bar.config.styleSettings.showBackground = value
				
				if self.bar.frames and self.bar.frames.background then
					if value then
						self.bar.frames.background:Show()
					else
						self.bar.frames.background:Hide()
					end
				end
			end,
		}
	end

	Factory = {}
	Factory.__index = Factory

	function Factory:New( metaName, metaStyle )
	    local self = {}
	    setmetatable( self, Factory )
	    
		self.metaName = metaName
		self.metaStyle = metaStyle or metaStyles[metaName]
		self.name = metaName
		self.title = metaStyle.title or metaName
		
		return self
	end

	local function copyTable( src, dest )
		if type( dest ) ~= "table" then dest = {} end
		if type( src ) == "table" then
			for k, v in pairs( src ) do
				if type( v ) == "table" then
					-- try to index the key first so that the metatable creates the defaults, if set, and use that table
					v = copyTable( v, dest[k] )
				end
				
				dest[k] = v
			end
		end
		
		return dest
	end

	function Factory:DefaultSettings( config )
	    local metaConfig = self.metaStyle.config
	    
	    if metaConfig then
	        if metaConfig.width then config.width = metaConfig.width end
	        if metaConfig.height then config.height = metaConfig.height end
		end

		if type( self.metaStyle.styleSettings ) == "table" then
		    config.styleSettings = copyTable( self.metaStyle.styleSettings )
		end
	end

	function Factory:Create( bar )
	    return Style:Create( bar, self.metaStyle )
	end
	
	for k, v in pairs( metaStyles ) do
	    local factory = Factory:New( k, v )
	    
	    UHB:RegisterStyle( factory )
	end
end
